----------------------------------------------------------------------------
-- LuaJIT SW64 disassembler module.
--
-- Copyright (C) 2019 deepin inc. All rights reserved.
-- Released under the MIT/X license. See Copyright Notice in luajit.h
----------------------------------------------------------------------------
-- This is a helper module used by the LuaJIT machine code dumper module.
--
-- It disassembles all standard SW64 instructions.
------------------------------------------------------------------------------
local type = type
local byte, format = string.byte, string.format
local match, gmatch = string.match, string.gmatch
local concat = table.concat
local bit = require("bit")
local band, bor, tohex = bit.band, bit.bor, bit.tohex
local lshift, rshift, arshift = bit.lshift, bit.rshift, bit.arshift

------------------------------------------------------------------------------
-- Primary and extended opcode maps
------------------------------------------------------------------------------

local OPC_SYSCALL = ""
local OPC_MISI_MEMORY, OPC_FUNC_MEMORY = "", ""
local OPC_MEMORY_F = "FBo"
local OPC_MEMORY = "ABo"
local OPC_CONTROL = "Ab"
local OPC_CONTROL_F = "Fb"
local OPC_ARITHMETIC = {
   shift = 5,
   mask = 0xff,
   pat = "ABD",
}
local OPC_ARITHMETIC_F = {
   shift = 5,
   mask = 0xff,
   pat = "FGI",
}
local OPC_ARITHMETIC_I = {
   shift = 5,
   mask = 0xff,
   pat = "AjD",
}
local OPC_COMPLEX_ARITHMETIC = {
   shift = 10,
   mask = 0x7,
   pat = "ABCD",
}
local OPC_COMPLEX_ARITHMETIC_F = {
   shift = 10,
   mask = 0x1f,
   pat = "FGHI",
}

local ignores_tabs = {
   F = { "FCVT", "IFMOV", "FCPY" },
}

function should_ignore(name, field)
   pat = ignores_tabs[field] or {}
   for _, p in ipairs(pat) do
      if match(name, p) then
         return true
      end
   end
   return false
end

local class_tabs = {
	[0x00] = OPC_SYSCALL,
	[0x01] = OPC_MEMORY,
	[0x02] = OPC_MEMORY,
	[0x03] = OPC_MEMORY,
	[0x04] = OPC_CONTROL,
	[0x05] = OPC_CONTROL,
	[0x06] = OPC_MISI_MEMORY,
	[0x08] = OPC_FUNC_MEMORY,
	[0x10] = OPC_ARITHMETIC,
	[0x11] = OPC_COMPLEX_ARITHMETIC,
	[0x12] = OPC_ARITHMETIC_I,
	[0x13] = OPC_ARITHMETIC_I,
	[0x18] = OPC_ARITHMETIC_F,
	[0x19] = OPC_COMPLEX_ARITHMETIC_F,
	[0x20] = OPC_MEMORY,
	[0x21] = OPC_MEMORY,
	[0x22] = OPC_MEMORY,
	[0x23] = OPC_MEMORY,
	[0x24] = OPC_MEMORY,
	[0x25] = OPC_MEMORY,
	[0x26] = OPC_MEMORY_F,
	[0x27] = OPC_MEMORY_F,
	[0x28] = OPC_MEMORY,
	[0x29] = OPC_MEMORY,
	[0x2A] = OPC_MEMORY,
	[0x2B] = OPC_MEMORY,
	[0x2C] = OPC_MEMORY,
	[0x2D] = OPC_MEMORY,
	[0x2E] = OPC_MEMORY_F,
	[0x2F] = OPC_MEMORY_F,
	[0x30] = OPC_CONTROL,
	[0x31] = OPC_CONTROL,
	[0x32] = OPC_CONTROL,
	[0x33] = OPC_CONTROL,
	[0x34] = OPC_CONTROL,
	[0x35] = OPC_CONTROL,
	[0x36] = OPC_CONTROL,
	[0x37] = OPC_CONTROL,
	[0x38] = OPC_CONTROL_F,
	[0x39] = OPC_CONTROL_F,
	[0x3A] = OPC_CONTROL_F,
	[0x3B] = OPC_CONTROL_F,
	[0x3C] = OPC_CONTROL_F,
	[0x3D] = OPC_CONTROL_F,
	[0x3e] = OPC_MEMORY,
	[0x3f] = OPC_MEMORY,
}

local map_pri = {
   [0x00] = {
      [0x0] = "SYSCALL/B",
      [0x1] = "SYSCALL"
   },
   [0x01]= {[0x0]= "CALL"},
   [0x02]= {[0x0]= "RET"},
   [0x03]= {[0x0]= "JMP"},
   [0x04]= {[0x0]= "BR"},
   [0x05]= {[0x0]= "BSR"},
   [0x06]= {
      [0x0000] = "MEMB",
      [0x0001] = "IMEMB",
      [0x1000] = "RD_F",
      [0x1020] = "WR_F",
   },
   [0x08]= {
      [0x0]= "LLDW",
      [0x1]= "LLDL",
      [0x8]= "LSTW",
      [0x9]= "LSTL",
	},
   [0x10]= {
		[0x00]= "ADDW",
		[0x01]= "SUBW",
		[0x02]= "S4ADDW",
		[0x03]= "S4SUBW",
		[0x04]= "S8ADDW",
		[0x05]= "S8SUBW",
		[0x08]= "ADDL",
		[0x09]= "SUBL",
		[0x0a]= "S4ADDL",
		[0x0b]= "S4SUBL",
		[0x0c]= "S8ADDL",
		[0x0d]= "S8SUBL",
		[0x10]= "MULW",
		[0x11]= "DIVW",
		[0x12]= "UDIVW",
		[0x13]= "REMW",
		[0x14]= "UREMW",
		[0x18]= "MULL",
		[0x19]= "UMULH",
	        [0x1a]= "DIVL",
		[0x1b]= "UDIVL",
		[0x1c]= "REML",
		[0x1d]= "UREML",
		[0x1e]= "ADDPI",
		[0x1f]= "ADDPIS",
		[0x28]= "CMPEQ",
		[0x29]= "CMPLT",
		[0x2a]= "CMPLE",
		[0x2b]= "CMPULT",
		[0x2c]= "CMPULE",
		[0x2d]= "SBT",
		[0x2e]= "CBT",
		[0x38]= "AND",
		[0x39]= "BIC",
		[0x3a]= "BIS",
		[0x3b]= "ORNOT",
		[0x3c]= "XOR",
		[0x3d]= "EQV",
		[0x40]= "INSLB",
		[0x41]= "INSLH",
		[0x42]= "INSLW",
		[0x43]= "INSLL",
		[0x44]= "INSHB",
		[0x45]= "INSHH",
		[0x46]= "INSHW",
		[0x47]= "INSHL",
		[0x48]= "SLL",
		[0x49]= "SRL",
		[0x4a]= "SRA",
		[0x4b]= "ROLL",
		[0x4c]= "SLLW",
		[0x4d]= "SRLW",
		[0x4e]= "SRAW",
		[0x4f]= "ROLW",
		[0x50]= "EXTLB",
		[0x51]= "EXTLH",
		[0x52]= "EXTLW",
		[0x53]= "EXTLL",
		[0x54]= "EXTHB",
		[0x55]= "EXTHH",
		[0x56]= "EXTHW",
		[0x57]= "EXTHL",
		[0x58]= "CTPOP",
		[0x59]= "CTLZ",
		[0x5a]= "CTTZ",
		[0x60]= "MASKLB",
		[0x61]= "MASKLH",
		[0x62]= "MASKLW",
		[0x63]= "MASKLL",
		[0x64]= "MASKHB",
		[0x65]= "MASKHH",
		[0x66]= "MASKHW",
		[0x67]= "MASKHL",
		[0x68]= "ZAP",
		[0x69]= "ZAPNOT",
		[0x6a]= "SEXTB",
		[0x6b]= "SEXTH",
		[0x6c]= "CMPGEB",
		[0x70]= "FIMOVS",
		[0x78]= "FIMOVD",
	},
   [0x11]= {
		[0x0]= "SELEQ",
		[0x1]= "SELGE",
		[0x2]= "SELGT",
		[0x3]= "SELLE",
		[0x4]= "SELLT",
		[0x5]= "SELNE",
		[0x6]= "SELLBC",
		[0x7]= "SELLBS",
	},
   [0x12]= {
		[0x00]= "ADDW",
		[0x01]= "SUBW",
		[0x02]= "S4ADDW",
		[0x03]= "S4SUBW",
		[0x04]= "S8ADDW",
		[0x05]= "S8SUBW",
		[0x08]= "ADDL",
		[0x09]= "SUBL",
		[0x0a]= "S4ADDL",
		[0x0b]= "S4SUBL",
		[0x0c]= "S8ADDL",
		[0x0d]= "S8SUBL",
		[0x10]= "MULW",
		[0x18]= "MULL",
		[0x19]= "UMULH",
		[0x28]= "CMPEQ",
		[0x29]= "CMPLT",
		[0x2a]= "CMPLE",
		[0x2b]= "CMPULT",
		[0x2c]= "CMPULE",
		[0X2d]= "SBT"
		[0x2e]= "CBT",
		[0x38]= "AND",
		[0x39]= "BIC",
		[0x3a]= "BIS",
		[0x3b]= "ORNOT",
		[0x3c]= "XOR",
		[0x3d]= "EQV",
		[0x40]= "INSLB",
		[0x41]= "INSLH",
		[0x42]= "INSLW",
		[0x43]= "INSLL",
		[0x44]= "INSHB",
		[0x45]= "INSHH",
		[0x46]= "INSHW",
		[0x47]= "INSHL",
		[0x48]= "SLL",
		[0x49]= "SRL",
		[0x4a]= "SRA",
		[0x4b]= "ROLL",
		[0x4c]= "SLLW",
		[0x4d]= "SRLW",
		[0x4e]= "SRAW",
		[0x4f]= "ROLW"
		[0x50]= "EXTLB",
		[0x51]= "EXTLH",
		[0x52]= "EXTLW",
		[0x53]= "EXTLL",
		[0x54]= "EXTHB",
		[0x55]= "EXTHH",
		[0x56]= "EXTHW",
		[0x57]= "EXTHL",
		[0x60]= "MASKLB",
		[0x61]= "MASKLH",
		[0x62]= "MASKLW",
		[0x63]= "MASKLL",
		[0x64]= "MASKHB",
		[0x65]= "MASKHH",
		[0x66]= "MASKHW",
		[0x67]= "MASKHL",
		[0x68]= "ZAP",
		[0x69]= "ZAPNOT",
		[0x6a]= "SEXTB",
		[0x6b]= "SEXTH",
		[0x6c]= "CMPGEB",
	},
    [0x13]= {
		[0x0]= "SELEQ",
		[0x1]= "SELGE",
		[0x2]= "SELGT",
		[0x3]= "SELLE",
		[0x4]= "SELLT",
		[0x5]= "SELNE",
		[0x6]= "SELLBC",
		[0x7]= "SELLBS",
    },
	[0x18]= {
		[0x00]= "FADDS",
		[0x01]= "FADDD",
		[0x02]= "FSUBS",
		[0x03]= "FSUBD",
		[0x04]= "FMULS",
		[0x05]= "FMULD",
		[0x06]= "FDIVS",
		[0x07]= "FDIVD",
		[0x08]= "FSQRTS",
		[0x09]= "FSQRTD",
		[0x10]= "FCMPEQ",
		[0x11]= "FCMPLE",
		[0x12]= "FCMPLT",
		[0x13]= "FCMPUN",
		[0x20]= "FCVTSD",
		[0x21]= "FCVTDS",
		[0x22]= "FCVTDL_G",
		[0x23]= "FCVTDL_P",
		[0x24]= "FCVTDL_Z",
		[0x25]= "FCVTDL_N",
		[0x27]= "FCVTDL",
		[0x28]= "FCVTWL",
		[0x29]= "FCVTLW",
		[0x2D]= "FCVTLS",
		[0x2F]= "FCVTLD",
		[0x30]= "FCPYS",
		[0x31]= "FCPYSE",
		[0x32]= "FCPYSN",
		[0x40]= "IFMOVS",
		[0x41]= "IFMOVD",
		[0x50]= "RFPCR",
		[0x51]= "WFPCR",
		[0x54]= "SETFPEC0",
		[0x55]= "SETFPEC1",
		[0x56]= "SETFPEC2",
		[0x57]= "SETFPEC3",
	},
    [0x19]= {
		[0x00]= "FMAS",
		[0x01]= "FMAD",
		[0x02]= "FMSS",
		[0x03]= "FMSD",
		[0x04]= "FNMAS",
		[0x05]= "FNMAD",
		[0x06]= "FNMSS",
		[0x07]= "FNMSD",

		[0x10]= "FSELEQ",
		[0x11]= "FSELNE",
		[0x12]= "FSELLT",
		[0x13]= "FSELLE",
		[0x14]= "FSELGT",
		[0x15]= "FSELGE",
	},
	[0x1D]= {[0]= "LBR"},
	[0x20]= {[0]= "LDBU"},
	[0x21]= {[0]= "LDHU"},
	[0x22]= {[0]= "LDW"},
	[0x23]= {[0]= "LDL"},
	[0x24]= {[0]= "LDL_U"},
	[0x25]= {[0]= "PRI_LD"},
	[0x26]= {[0]= "FLDS"},
	[0x27]= {[0]= "FLDD"},
	[0x28]= {[0]= "STB"},
	[0x29]= {[0]= "STH"},
	[0x2A]= {[0]= "STW"},
	[0x2B]= {[0]= "STL"},
	[0x2C]= {[0]= "STL_U"},
	[0x2D]= {[0]= "PRI_ST"},
	[0x2E]= {[0]= "FSTS"},
	[0x2F]= {[0]= "FSTD"},
	[0x30]= {[0]= "BEQ"},
	[0x31]= {[0]= "BNE"},
	[0x32]= {[0]= "BLT"},
	[0x33]= {[0]= "BLE"},
	[0x34]= {[0]= "BGT"},
	[0x35]= {[0]= "BGE"},
	[0x36]= {[0]= "BLBC"},
	[0x37]= {[0]= "BLBS"},
	[0x38]= {[0]= "FBEQ"},
	[0x39]= {[0]= "FBNE"},
	[0x3A]= {[0]= "FBLT"},
	[0x3B]= {[0]= "FBLE"},
	[0x3C]= {[0]= "FBGT"},
	[0x3D]= {[0]= "FBGE"},
	[0x3e]= {[0]= "LDI"},
	[0x3f]= {[0]= "LDIH"},
}


------------------------------------------------------------------------------

local map_gpr = {
   [0] = "r0", "r1", "r2", "r3", "r4", "r5", "r6", "r7",
   "r8", "BASE", "r10", "r11", "r12", "r13", "r14", "JGL",
   "r16", "r17", "r18", "r19", "r20", "r21", "r22", "r23",
   "r24", "r25", "JTMP", "r27", "at", "r29", "sp", "zero",
}

------------------------------------------------------------------------------

-- Output a nicely formatted line with an opcode and operands.
local function putop(ctx, text, operands)
   local pos = ctx.pos
   local extra = ""
   if ctx.rel then
      local sym = ctx.symtab[ctx.rel]
      if sym then extra = "\t->"..sym end
   end
   if ctx.hexdump > 0 then
      ctx.out(format("%08x  %s  %-7s %s%s\n",
                     ctx.addr+pos, tohex(ctx.op), text, concat(operands, ", "), extra))
   else
      ctx.out(format("%08x  %-7s %s%s\n",
                     ctx.addr+pos, text, concat(operands, ", "), extra))
   end
   ctx.pos = pos + 4
end

-- Fallback for unknown opcodes.
local function unknown(ctx)
   return putop(ctx, ".long", { "0x"..tohex(ctx.op) })
end

local function get_le(ctx)
   local pos = ctx.pos
   local b0, b1, b2, b3 = byte(ctx.code, pos+1, pos+4)
   return bor(lshift(b3, 24), lshift(b2, 16), lshift(b1, 8), b0)
end

-- Disassemble a single instruction.
local function disass_ins(ctx)
   local op = ctx:get()
   local operands = {}
   local last = nil
   ctx.op = op
   ctx.rel = nil

   local opcode = band(rshift(op, 26), 0x3f)
   local opat = map_pri[opcode]

   local fn = 0
   local pat = class_tabs[opcode]
   local name = opat[0]
   if pat.shift then
      name = opat[band(rshift(op, pat.shift), pat.mask)]
      pat = pat.pat
   end
   local isf = false

   if name == "FIMOVD" then
      pat = "FD"
   elseif name == "IFMOVD" then
      pat = "AI"
   end


   for p in gmatch(pat, ".") do
      local x = nil
      if should_ignore(name, p) then
         -- do nothing
      elseif p == "A" then
         x = map_gpr[band(rshift(op, 21), 31)]
      elseif p == "B" then
         x = map_gpr[band(rshift(op, 16), 31)]
      elseif p == "C" then
         x = map_gpr[band(rshift(op, 5), 31)]
      elseif p == "D" then
         x = map_gpr[band(rshift(op, 0), 31)]
      elseif p == "F" then
         isf = true
         x = "f"..band(rshift(op, 21), 31)
      elseif p == "G" then
         isf = true
         x = "f"..band(rshift(op, 16), 31)
      elseif p == "H" then
         isf = true
         x = "f"..band(rshift(op, 5), 31)
      elseif p == "I" then
         isf = true
         x = "f"..band(rshift(op, 0), 31)
      elseif p == "o" then
         local disp = arshift(lshift(op, 16), 16)
         if name == "LDI" and disp == 0 then
            name = "MOVE"
            operands[#operands] = last
         else
            operands[#operands] = format("%d(%s)", disp, last)
         end
      elseif p == "p" then
         local index = map_gpr[band(rshift(op, 16), 31)]
         operands[#operands] = format("%s(%s)", index, last)
      elseif p == "b" then
         x = ctx.addr + ctx.pos + arshift(lshift(op, 21), 21)*4 + 4
         ctx.rel = x
         x = format("0x%08x", x)
      elseif p == "i" then
         x = band(rshift(op, 13), 0x000ff)
      elseif p == "j" then
         x = band(rshift(op, 13), 0x000ff)
      elseif p == "j" then
         x = band(rshift(op, 13), 0x000ff)
      elseif p == "1" then
         if last == "ra" then
            operands[#operands] = nil
         end
      else
         assert(false)
      end
      if x then operands[#operands+1] = x; last = x end
   end
   return putop(ctx, name, operands)
end

------------------------------------------------------------------------------

-- Disassemble a block of code.
local function disass_block(ctx, ofs, len)
   if not ofs then ofs = 0 end
   local stop = len and ofs+len or #ctx.code
   stop = stop - stop % 4
   ctx.pos = ofs - ofs % 4
   ctx.rel = nil
   while ctx.pos < stop do disass_ins(ctx) end
end

-- Extended API: create a disassembler context. Then call ctx:disass(ofs, len).
local function create(code, addr, out)
   local ctx = {}
   ctx.code = code
   ctx.addr = addr or 0
   ctx.out = out or io.write
   ctx.symtab = {}
   ctx.disass = disass_block
   ctx.hexdump = 8
   ctx.get = get_le
   return ctx
end

-- Simple API: disassemble code (a string) at address and output via out.
local function disass(code, addr, out)
   create(code, addr, out):disass()
end

-- Return register name for RID.
local function regname(r)
   if r < 32 then return map_gpr[r] end
   return "f"..(r-32)
end



function wi_debug(__obj, op, addr)
   if not addr then
      addr = 0
   end
   local operands = {}
   local last = nil

   local opcode = band(rshift(op, 26), 0x3f)
   local opat = map_pri[opcode]

   local fn = 0
   local pat = class_tabs[opcode]
   local name = opat[0]
   if pat.shift then
      name = opat[band(rshift(op, pat.shift), pat.mask)]
      pat = pat.pat
   end
   local isf = false

   if name == "FIMOVD" then
      pat = "FD"
   elseif name == "IFMOVD" then
      pat = "AI"
   end

   for p in gmatch(pat, ".") do
      local x = nil
      if should_ignore(name, p) then
         -- do nothing
      elseif p == "A" then
         x = map_gpr[band(rshift(op, 21), 31)]
      elseif p == "B" then
         x = map_gpr[band(rshift(op, 16), 31)]
      elseif p == "C" then
         x = map_gpr[band(rshift(op, 5), 31)]
      elseif p == "D" then
         x = map_gpr[band(rshift(op, 0), 31)]
      elseif p == "F" then
         isf = true
         x = "f"..band(rshift(op, 21), 31)
      elseif p == "G" then
         isf = true
         x = "f"..band(rshift(op, 16), 31)
      elseif p == "H" then
         isf = true
         x = "f"..band(rshift(op, 5), 31)
      elseif p == "I" then
         isf = true
         x = "f"..band(rshift(op, 0), 31)
      elseif p == "o" then
         local disp = arshift(lshift(op, 16), 16)
         if name == "LDI" and disp == 0 then
            name = "MOVE"
            operands[#operands] = last
         else
            operands[#operands] = format("%d(%s)", disp, last)
         end
      elseif p == "p" then
         local index = map_gpr[band(rshift(op, 16), 31)]
         operands[#operands] = format("%s(%s)", index, last)
      elseif p == "b" then
         x = addr + arshift(lshift(op, 21), 21)*4 + 4
         x = format("0x%08x", x)
      elseif p == "i" then
         x = band(rshift(op, 13), 0x000ff)
      elseif p == "j" then
         x = band(rshift(op, 13), 0x000ff)
      elseif p == "j" then
         x = band(rshift(op, 13), 0x000ff)
      elseif p == "1" then
         if last == "ra" then
            operands[#operands] = nil
         end
      else
         assert(false)
      end
      if x then operands[#operands+1] = x; last = x end
   end
   print(name, concat(operands, ", "))
end

-- Public module functions.
return {
   create = create,
   disass = disass,
   regname = regname,
   wi_debug = wi_debug,
}
